==> js-20.0.0-sources.jar: com/oracle/truffle/js/runtime/objects/JSSharedData.java <==
/*
 * Copyright (c) 2018, 2018, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * The Universal Permissive License (UPL), Version 1.0
 *
 * Subject to the condition set forth below, permission is hereby granted to any
 * person obtaining a copy of this software, associated documentation and/or
 * data (collectively the "Software"), free of charge and under any and all
 * copyright rights in the Software, and any and all patent rights owned or
 * freely licensable by each licensor hereunder covering either (i) the
 * unmodified Software as contributed to or provided by such licensor, or (ii)
 * the Larger Works (as defined below), to deal in both
 *
 * (a) the Software, and
 *
 * (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if
 * one is included with the Software each a "Larger Work" to which the Software
 * is contributed by such licensors),
 *
 * without restriction, including without limitation the rights to copy, create
 * derivative works of, display, perform, and distribute the Software and make,
 * use, sell, offer for sale, import, export, have made, and have sold the
 * Software and the Larger Work(s), and to sublicense the foregoing rights on
 * either these or other terms.
 *
 * This license is subject to the following condition:
 *
 * The above copyright notice and either this complete permission notice or at a
 * minimum a reference to the UPL must be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.oracle.truffle.js.runtime.objects;

import org.graalvm.collections.EconomicMap;

import com.oracle.truffle.api.Assumption;
import com.oracle.truffle.api.CompilerAsserts;
import com.oracle.truffle.api.Truffle;
import com.oracle.truffle.api.object.Property;
import com.oracle.truffle.api.object.ShapeListener;
import com.oracle.truffle.api.utilities.NeverValidAssumption;
import com.oracle.truffle.js.runtime.JSContext;
import com.oracle.truffle.js.runtime.util.DebugCounter;

/**
 * @see JSShape
 */
public final class JSSharedData implements ShapeListener {
    private final JSContext context;
    private final Property prototypeProperty;
    private EconomicMap<Object, Assumption> propertyAssumptions;
    private Assumption prototypeAssumption;

    private static final DebugCounter propertyAssumptionsCreated = DebugCounter.create("Property assumptions created");
    private static final DebugCounter propertyAssumptionsRemoved = DebugCounter.create("Property assumptions removed");
    private static final DebugCounter prototypeAssumptionsCreated = DebugCounter.create("Prototype assumptions created");
    private static final DebugCounter prototypeAssumptionsRemoved = DebugCounter.create("Prototype assumptions removed");

    public JSSharedData(JSContext context, Property prototypeProperty) {
        this.context = context;
        this.prototypeProperty = prototypeProperty;
    }

    synchronized Assumption getPropertyAssumption(Object propertyName) {
        CompilerAsserts.neverPartOfCompilation();
        EconomicMap<Object, Assumption> map = propertyAssumptions;
        if (map == null) {
            map = EconomicMap.create();
            propertyAssumptions = map;
        } else {
            Assumption assumption = map.get(propertyName);
            if (assumption != null) {
                return assumption;
            }
        }
        Assumption assumption = Truffle.getRuntime().createAssumption(propertyName.toString());
        map.put(propertyName, assumption);
        propertyAssumptionsCreated.inc();
        return assumption;
    }

    synchronized void invalidatePropertyAssumption(Object propertyName) {
        CompilerAsserts.neverPartOfCompilation();
        EconomicMap<Object, Assumption> map = propertyAssumptions;
        if (map != null) {
            Assumption assumption = map.get(propertyName);
            if (assumption != null && assumption != NeverValidAssumption.INSTANCE) {
                assumption.invalidate("invalidatePropertyAssumption");
                map.put(propertyName, NeverValidAssumption.INSTANCE);
                propertyAssumptionsRemoved.inc();
            }
        }
    }

    JSContext getContext() {
        return context;
    }

    Property getPrototypeProperty() {
        return prototypeProperty;
    }

    synchronized Assumption getPrototypeAssumption() {
        Assumption assumption = prototypeAssumption;
        if (assumption == null) {
            assumption = Truffle.getRuntime().createAssumption("stable prototype");
            prototypeAssumption = assumption;
            prototypeAssumptionsCreated.inc();
        }
        return assumption;
    }

    synchronized void invalidatePrototypeAssumption() {
        Assumption assumption = prototypeAssumption;
        if (assumption != null) {
            assumption.invalidate();
            prototypeAssumption = NeverValidAssumption.INSTANCE;
            prototypeAssumptionsRemoved.inc();
        }
    }

    @Override
    public void onPropertyTransition(Object key) {
        invalidatePropertyAssumption(key);
    }
}

==> js-22.0.0.2-sources.jar: com/oracle/truffle/js/runtime/objects/JSSharedData.java <==
/*
 * Copyright (c) 2018, 2020, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * The Universal Permissive License (UPL), Version 1.0
 *
 * Subject to the condition set forth below, permission is hereby granted to any
 * person obtaining a copy of this software, associated documentation and/or
 * data (collectively the "Software"), free of charge and under any and all
 * copyright rights in the Software, and any and all patent rights owned or
 * freely licensable by each licensor hereunder covering either (i) the
 * unmodified Software as contributed to or provided by such licensor, or (ii)
 * the Larger Works (as defined below), to deal in both
 *
 * (a) the Software, and
 *
 * (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if
 * one is included with the Software each a "Larger Work" to which the Software
 * is contributed by such licensors),
 *
 * without restriction, including without limitation the rights to copy, create
 * derivative works of, display, perform, and distribute the Software and make,
 * use, sell, offer for sale, import, export, have made, and have sold the
 * Software and the Larger Work(s), and to sublicense the foregoing rights on
 * either these or other terms.
 *
 * This license is subject to the following condition:
 *
 * The above copyright notice and either this complete permission notice or at a
 * minimum a reference to the UPL must be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.oracle.truffle.js.runtime.objects;

import com.oracle.truffle.api.Assumption;
import com.oracle.truffle.api.Truffle;
import com.oracle.truffle.api.utilities.NeverValidAssumption;
import com.oracle.truffle.js.runtime.JSContext;
import com.oracle.truffle.js.runtime.util.DebugCounter;

/**
 * @see JSShape
 */
public final class JSSharedData {
    private final JSContext context;
    private final JSDynamicObject proto;
    private Assumption prototypeAssumption;

    private static final DebugCounter prototypeAssumptionsCreated = DebugCounter.create("Prototype assumptions created");
    private static final DebugCounter prototypeAssumptionsRemoved = DebugCounter.create("Prototype assumptions removed");

    public JSSharedData(JSContext context, JSDynamicObject prototype) {
        this.context = context;
        this.proto = prototype;
    }

    JSContext getContext() {
        return context;
    }

    public JSDynamicObject getPrototype() {
        return proto;
    }

    synchronized Assumption getPrototypeAssumption() {
        Assumption assumption = prototypeAssumption;
        if (assumption == null) {
            assumption = Truffle.getRuntime().createAssumption("stable prototype");
            prototypeAssumption = assumption;
            prototypeAssumptionsCreated.inc();
        }
        return assumption;
    }

    synchronized void invalidatePrototypeAssumption() {
        Assumption assumption = prototypeAssumption;
        if (assumption != null) {
            assumption.invalidate();
            prototypeAssumption = NeverValidAssumption.INSTANCE;
            prototypeAssumptionsRemoved.inc();
        }
    }
}
